@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2019 Artifex Software, Inc.
+/* Copyright (C) 2001-2023 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -9,8 +9,8 @@
    of the license contained in the file LICENSE in this distribution.
 
    Refer to licensing information at http://www.artifex.com or contact
-   Artifex Software, Inc.,  1305 Grant Avenue - Suite 200, Novato,
-   CA 94945, U.S.A., +1(415)492-9861, for further information.
+   Artifex Software, Inc.,  39 Mesa Street, Suite 108A, San Francisco,
+   CA 94129, USA, for further information.
 */
 
 
@@ -102,7 +102,7 @@ build_foreground(pcl_state_t * pcs,
     }
 
     num_entries = pindexed->num_entries;
-    
+
     /* release the existing foreground */
     if (pfrgrnd != 0) {
         rc_decrement(pfrgrnd, "build pcl foreground");
@@ -201,6 +201,8 @@ static int
 frgrnd_do_reset(pcl_state_t * pcs, pcl_reset_type_t type)
 {
     if (type & (pcl_reset_permanent)) {
+        /* pcl_free_default_objects() makes similar calls; probably need to
+        keep synced. */
         rc_decrement(pcs->pfrgrnd, "foreground reset pfrgrnd");
         rc_decrement(pcs->pdflt_frgrnd, "foreground reset pdflt_frgrnd");
         rc_decrement(pcs->pwhite_cs, "foreground reset p_white_cs");
