@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2018 Artifex Software, Inc.
+/* Copyright (C) 2001-2025 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -9,8 +9,8 @@
    of the license contained in the file LICENSE in this distribution.
 
    Refer to licensing information at http://www.artifex.com or contact
-   Artifex Software, Inc.,  1305 Grant Avenue - Suite 200, Novato,
-   CA 94945, U.S.A., +1(415)492-9861, for further information.
+   Artifex Software, Inc.,  39 Mesa Street, Suite 108A, San Francisco,
+   CA 94129, USA, for further information.
 */
 
 
@@ -23,10 +23,12 @@
 #include "scommon.h"
 #include "gdebug.h"
 #include "gp.h"
+#include "gpmisc.h"
 #include "gxiodev.h"
 #include "pjparse.h"
 #include "plfont.h"
 #include "plver.h"              /* PJL_VOLUME_0 PJL_VOLUME_1 */
+#include "plmain.h"
 #include "ctype_.h"             /* for toupper() */
 #include <stdlib.h>             /* for atoi() */
 
@@ -82,7 +84,7 @@ typedef struct pjl_parser_state_s
     int line_size;
     int bytes_to_write;         /* data processing by fsdownload */
     int bytes_to_read;          /* data processed by fsupload */
-    FILE *fp;                   /* fsdownload or fsupload file */
+    gp_file *fp;                   /* fsdownload or fsupload file */
     int pos;                    /* current position in line */
     pjl_envir_var_t *defaults;  /* the default environment (i.e. set default) */
     pjl_envir_var_t *envir;     /* the pjl environment */
@@ -92,6 +94,10 @@ typedef struct pjl_parser_state_s
     char *environment_font_path;        /* if there is an operating sytem env
                                            var it is used instead of the
                                            default pjl fontsource */
+    /* permenant soft font slots - bit n is the n'th font number. */
+    #define MAX_PERMANENT_FONTS 256 /* multiple of 8 */
+    unsigned char permanent_soft_fonts[MAX_PERMANENT_FONTS / 8];
+
     gs_memory_t *mem;
 } pjl_parser_state_t;
 
@@ -193,7 +199,9 @@ typedef enum
     NAME,                       /* used for pathnames */
     SIZE,                       /* size of data */
     VOLUME,                     /* volume indicator for filesystem initialization */
-    DISKLOCK
+    DISKLOCK,
+    GSSET,
+    GSSETSTRING
 } pjl_token_type_t;
 
 /* lookup table to map strings to pjl tokens */
@@ -230,6 +238,8 @@ static const pjl_lookup_table_t pjl_table[] = {
     {"COUNT", COUNT},
     {"OFFSET", OFFSET},
     {"DISKLOCK", DISKLOCK},
+    {"GSSET", GSSET},
+    {"GSSETSTRING", GSSETSTRING},
     {"", (pjl_token_type_t) 0 /* don't care */ }
 };
 
@@ -245,10 +255,6 @@ static const struct
     px_enumerate_media(PJLMEDIA)
 };
 
-/* permenant soft font slots - bit n is the n'th font number. */
-#define MAX_PERMANENT_FONTS 256 /* multiple of 8 */
-unsigned char pjl_permanent_soft_fonts[MAX_PERMANENT_FONTS / 8];
-
 /* ----- private functions and definitions ------------ */
 
 /* forward declaration */
@@ -319,10 +325,10 @@ pjl_side_effects(pjl_parser_state_t * pst, char *variable, char *value,
                            (int)(pjl_media[indx].width) :
                            (int)(pjl_media[indx].height));
         int formlines = pjl_calc_formlines_new_page_size(page_length);
-        char value[32];
+        char text[32];
 
-        gs_sprintf(value, "%d", formlines);
-        pjl_set(pst, (char *)"FORMLINES", value, defaults);
+        gs_snprintf(text, sizeof(text), "%d", formlines);
+        pjl_set(pst, (char *)"FORMLINES", text, defaults);
     }
     /* fill in other side effects here */
     return;
@@ -439,31 +445,35 @@ pjl_check_font_path(char *path_list, gs_memory_t * mem)
 {
     /* lookup a font path and check if any files (presumably fonts are
        present) */
-    char tmp_path[PJL_PATH_NAME_LENGTH + 1];
-    char *tmp_pathp = tmp_path, *tplast = NULL;
+    char path[PJL_PATH_NAME_LENGTH + 1];
+    char *pathp = path, *tplast = NULL;
     const char pattern[] = "*";
-    char tmp_path_and_pattern[PJL_PATH_NAME_LENGTH + 1 + 1];    /* pattern + null */
+    char path_and_pattern[PJL_PATH_NAME_LENGTH + 1 + 1];    /* pattern + null */
     char *dirname;
     char fontfilename[MAXPATHLEN + 1];
 
-    /* make a tmp copy of the colon delimited path */
-    strcpy(tmp_path, path_list);
+    /* Ignore error if the path list is too long, a single path may be
+       okay. */
+    gs_strlcpy(path, path_list, sizeof(path));
     /* for each path search for fonts.  If we find them return we only
        check if the directory resource has files without checking if
        the files are indeed fonts. */
-    while ((dirname = gs_strtok(tmp_pathp, ";", &tplast)) != NULL) {
+    while ((dirname = gs_strtok(pathp, ";", &tplast)) != NULL) {
         file_enum *fe;
 
-        strcpy(tmp_path_and_pattern, dirname);
-        strcat(tmp_path_and_pattern, pattern);
+        if (gs_strlcpy(path_and_pattern, dirname, sizeof(path_and_pattern)) >= sizeof(path_and_pattern))
+            continue;
 
-        fe = gs_enumerate_files_init(tmp_path_and_pattern,
-                                     strlen(tmp_path_and_pattern), mem);
+        if (gs_strlcat(path_and_pattern, pattern, sizeof(path_and_pattern)) >= sizeof(path_and_pattern))
+            continue;
+
+        fe = gs_enumerate_files_init(mem, path_and_pattern,
+                                     strlen(path_and_pattern));
         if (fe == NULL
             ||
-            (gs_enumerate_files_next(fe, fontfilename, PJL_PATH_NAME_LENGTH))
+            (gs_enumerate_files_next(mem, fe, fontfilename, PJL_PATH_NAME_LENGTH))
             == -1) {
-            tmp_pathp = NULL;
+            pathp = NULL;
         } else {
             /* wind through the rest of the files.  This should close
                things up as well.  All we need to do is clean up but
@@ -471,16 +481,21 @@ pjl_check_font_path(char *path_list, gs_memory_t * mem)
                directory */
             while (1) {
                 int fstatus =
-                    (int)gs_enumerate_files_next(fe, fontfilename,
+                    (int)gs_enumerate_files_next(mem, fe, fontfilename,
                                                  PJL_PATH_NAME_LENGTH);
                 /* we don't care if the file does not fit (return +1) */
                 if (fstatus == -1)
                     break;
             }
-            /* NB fix me - replace : separated path with real path.
-               We should do this elsewhere */
-            strcpy(path_list, dirname);
-            return path_list;
+            /* replace : separated path with real path.  The path list
+               is of equal or greater length than dirname, so it can't
+               fail, but we check anyway. */
+            if (strlen(path_list) < strlen(dirname))
+                return NULL;
+            else {
+                strcpy(path_list, dirname);
+                return path_list;
+            }
         }
     }
     return NULL;
@@ -528,21 +543,20 @@ pjl_reset_fontsource_fontnumbers(pjl_parser_state_t * pst)
 static void
 pjl_parsed_filename_to_string(char *fnamep, const char *pathname)
 {
-    int i;
-    int size;
+    size_t i, size, prefix_size;
+    const char *prefix;
+    char name[MAXPATHLEN];
+    uint rlen;
 
     *fnamep = 0;                /* in case of bad input */
-    if (pathname == 0 || pathname[0] != '"' || strlen(pathname) < 3)
+    if (pathname == 0 || strlen(pathname) < 4 ||
+        pathname[0] != '"' || pathname[strlen(pathname)-1] != '"')
         return;                 /* bad input pjl file */
 
     if (pathname[1] == '0' && pathname[2] == ':') {
-        /* copy pjl_volume string in */
-        strncpy(fnamep, PJL_VOLUME_0, strlen(PJL_VOLUME_0));
-        fnamep += strlen(PJL_VOLUME_0);
+        prefix = PJL_VOLUME_0;
     } else if (pathname[1] == '1' && pathname[2] == ':') {
-        /* copy pjl_volume string in */
-        strncpy(fnamep, PJL_VOLUME_1, strlen(PJL_VOLUME_1));
-        fnamep += strlen(PJL_VOLUME_1);
+        prefix = PJL_VOLUME_1;
     } else
         return;                 /* bad input pjl file */
 
@@ -550,23 +564,35 @@ pjl_parsed_filename_to_string(char *fnamep, const char *pathname)
      * remove quotes, use forward slash, copy rest.
      */
     size = strlen(pathname);
+    if (size - 4 > MAXPATHLEN)
+        return;
+    for (i = 3; i < size - 1; i++)
+        name[i - 3] = (pathname[i] == '\\') ? '/' : pathname[i];
+    size -= 4;
 
-    for (i = 3; i < size; i++) {
-        if (pathname[i] == '\\')
-            *fnamep++ = '/';
-        else if (pathname[i] != '"')
-            *fnamep++ = pathname[i];
-        /* else it is a quote skip it */
-    }
-    /* NULL terminate */
-    *fnamep = '\0';
+    /* copy pjl_volume string in. use strlen+1 to ensure that strncpy()
+    appends '\0', to keep coverity quiet. */
+    prefix_size = strlen(prefix);
+    strncpy(fnamep, prefix, prefix_size + 1);
+
+    if (size == 0)
+        return;
+
+    if (name[0] != '/')
+        fnamep[prefix_size++] = '/';
+
+    rlen = MAXPATHLEN - prefix_size;
+    if (gp_file_name_reduce(name, size, fnamep + prefix_size, &rlen) == gp_combine_success)
+        fnamep[prefix_size + rlen] = 0;
+    else
+        fnamep[0] = 0;
 }
 
 /* Verify a file write operation is ok.  The filesystem must be 0: or
    1:, no other pjl files can have pending writes, and the pjl
    disklock state variable must be false */
 static int
-pjl_verify_file_operation(pjl_parser_state_t * pst, char *fname)
+pjl_verify_file_operation(pjl_parser_state_t * pst, char *fname, const char *access)
 {
     /* make sure we are playing in the pjl sandbox */
     if (0 != strncmp(PJL_VOLUME_0, fname, strlen(PJL_VOLUME_0))
@@ -574,6 +600,12 @@ pjl_verify_file_operation(pjl_parser_state_t * pst, char *fname)
         dmprintf1(pst->mem, "illegal path name %s\n", fname);
         return -1;
     }
+
+    if (access != NULL && gp_validate_path(pst->mem, fname, access) != 0) {
+        dmprintf1(pst->mem, "illegal path name %s\n", fname);
+        return -1;
+    }
+
     /* make sure we are not currently writing to a file.
        Simultaneously file writing is not supported */
     if (pst->bytes_to_write || pst->fp)
@@ -590,25 +622,25 @@ pjl_verify_file_operation(pjl_parser_state_t * pst, char *fname)
 static void
 pjl_warn_exists(const gs_memory_t * mem, char *fname)
 {
-    FILE *fpdownload;
+    gp_file *fpdownload;
 
     /* issue a warning if the file exists */
-    if ((fpdownload = gp_fopen(fname, gp_fmode_rb)) != NULL) {
-        fclose(fpdownload);
+    if ((fpdownload = gp_fopen(mem, fname, gp_fmode_rb)) != NULL) {
+        gp_fclose(fpdownload);
         dmprintf1(mem, "warning file exists overwriting %s\n", fname);
     }
 }
 
 /* open a file for writing or appending */
-static FILE *
+static gp_file *
 pjl_setup_file_for_writing(pjl_parser_state_t * pst, char *pathname, int size,
                            bool append)
 {
-    FILE *fp;
+    gp_file *fp;
     char fname[MAXPATHLEN];
 
     pjl_parsed_filename_to_string(fname, pathname);
-    if (pjl_verify_file_operation(pst, fname) < 0)
+    if (pjl_verify_file_operation(pst, fname, NULL) < 0)
         return NULL;
     pjl_warn_exists(pst->mem, fname);
     {
@@ -617,7 +649,7 @@ pjl_setup_file_for_writing(pjl_parser_state_t * pst, char *pathname, int size,
         strcpy(fmode, gp_fmode_wb);
         if (append)
             strcat(fmode, "+");
-        if ((fp = gp_fopen(fname, gp_fmode_wb)) == NULL) {
+        if ((fp = gp_fopen(pst->mem, fname, gp_fmode_wb)) == NULL) {
             dmprintf(pst->mem, "warning file open for writing failed\n");
             return NULL;
         }
@@ -630,7 +662,7 @@ static int
 pjl_set_fs_download_state(pjl_parser_state_t * pst, char *pathname, int size)
 {
     /* somethink is wrong if the state indicates we are already writing to a file */
-    FILE *fp =
+    gp_file *fp =
         pjl_setup_file_for_writing(pst, pathname, size, false /* append */ );
     if (fp == NULL)
         return -1;
@@ -644,7 +676,7 @@ pjl_set_fs_download_state(pjl_parser_state_t * pst, char *pathname, int size)
 static int
 pjl_set_append_state(pjl_parser_state_t * pst, char *pathname, int size)
 {
-    FILE *fp =
+    gp_file *fp =
         pjl_setup_file_for_writing(pst, pathname, size, true /* append */ );
     if (fp == NULL)
         return -1;
@@ -661,12 +693,16 @@ pjl_fsinit(pjl_parser_state_t * pst, char *pathname)
     char fname[MAXPATHLEN];
 
     pjl_parsed_filename_to_string(fname, pathname);
-    if (pjl_verify_file_operation(pst, fname) < 0)
+    if (pjl_verify_file_operation(pst, fname, "c") < 0)
         return -1;
 #ifdef GS_NO_FILESYSTEM
     return -1;
 #else
+  #ifdef __WIN32__
+    return _mkdir(fname);
+  #else
     return mkdir(fname, 0777);
+  #endif
 #endif
 }
 
@@ -677,12 +713,16 @@ pjl_fsmkdir(pjl_parser_state_t * pst, char *pathname)
     char fname[MAXPATHLEN];
 
     pjl_parsed_filename_to_string(fname, pathname);
-    if (pjl_verify_file_operation(pst, fname) < 0)
+    if (pjl_verify_file_operation(pst, fname, "c") < 0)
         return -1;
 #ifdef GS_NO_FILESYSTEM
     return -1;
 #else
+  #ifdef __WIN32__
+    return _mkdir(fname);
+  #else
     return mkdir(fname, 0777);
+  #endif
 #endif
 }
 
@@ -716,14 +756,12 @@ pjl_search_for_file(pjl_parser_state_t * pst, char *pathname, char *filename,
     struct stat stbuf;
 
     /* should check length */
-    strcpy(fontfilename, pathname);
-    strcat(fontfilename, "/*");
-    fe = gs_enumerate_files_init(fontfilename, strlen(fontfilename),
-                                 pst->mem);
+    snprintf(fontfilename, sizeof(fontfilename), "%s/*", pathname);
+    fe = gs_enumerate_files_init(pst->mem, fontfilename, strlen(fontfilename));
     if (fe) {
         do {
             uint fstatus =
-                gs_enumerate_files_next(fe, fontfilename,
+                gs_enumerate_files_next(pst->mem, fe, fontfilename,
                                         PJL_PATH_NAME_LENGTH);
             /* done */
             if (fstatus == ~(uint) 0)
@@ -748,17 +786,19 @@ static int
 pjl_fsdirlist(pjl_parser_state_t * pst, char *pathname, int entry, int count)
 {
     file_enum *fe;
-    char fontfilename[MAXPATHLEN];
+    char fontfilename[MAXPATHLEN + 2];
 
     pjl_parsed_filename_to_string(fontfilename, pathname);
+    if (pjl_verify_file_operation(pst, fontfilename, NULL) < 0)
+        return -1;
+
     /* if this is a directory add * for the directory listing NB fix */
     strcat(fontfilename, "/*");
-    fe = gs_enumerate_files_init(fontfilename, strlen(fontfilename),
-                                 pst->mem);
+    fe = gs_enumerate_files_init(pst->mem, fontfilename, strlen(fontfilename));
     if (fe) {
         do {
             uint fstatus =
-                gs_enumerate_files_next(fe, fontfilename,
+                gs_enumerate_files_next(pst->mem, fe, fontfilename,
                                         PJL_PATH_NAME_LENGTH);
             /* done */
             if (fstatus == ~(uint) 0)
@@ -781,15 +821,15 @@ pjl_write_remaining_data(pjl_parser_state_t * pst, const byte ** pptr,
     uint avail = plimit - ptr;
     uint bytes_written = min(avail, pst->bytes_to_write);
 
-    if (fwrite(ptr, 1, bytes_written, pst->fp) != bytes_written) {
+    if (gp_fwrite(ptr, 1, bytes_written, pst->fp) != bytes_written) {
         /* try to close the file before failing */
-        fclose(pst->fp);
+        gp_fclose(pst->fp);
         pst->fp = NULL;
         return -1;
     }
     pst->bytes_to_write -= bytes_written;
     if (pst->bytes_to_write == 0) {     /* done */
-        fclose(pst->fp);
+        gp_fclose(pst->fp);
         pst->fp = NULL;
     }
     /* update stream pointer */
@@ -803,9 +843,9 @@ pjl_delete_file(pjl_parser_state_t * pst, char *pathname)
     char fname[MAXPATHLEN];
 
     pjl_parsed_filename_to_string(fname, pathname);
-    if (pjl_verify_file_operation(pst, fname) < 0)
+    if (pjl_verify_file_operation(pst, fname, NULL) < 0)
         return -1;
-    return unlink(fname);
+    return gp_unlink(pst->mem, fname);
 }
 
 /* handle pattern foo = setting, e.g. volume = "0:", name = 0:]pcl.
@@ -825,6 +865,33 @@ pjl_get_setting(pjl_parser_state_t * pst, pjl_token_type_t tok, char *token)
     return 0;
 }
 
+static int
+gs_set(pjl_parser_state_t *pst, const char *variable, const char *token, int string)
+{
+    int code;
+    int len1 = strlen(variable)+1;
+    int len2 = NULL ? 0 : strlen(token)+1;
+    char *buffer = (char *)gs_alloc_bytes(pst->mem, len1+len2, "gs_set buffer");
+
+    if (buffer == NULL)
+        return -1;
+
+    strcpy(buffer, variable);
+    if (len2) {
+        strcat(buffer, "=");
+        strcat(buffer, token);
+    }
+
+    if (string)
+        code = pl_main_set_string_param(pl_main_get_instance(pst->mem), buffer);
+    else
+        code = pl_main_set_param(pl_main_get_instance(pst->mem), buffer);
+    gs_free_object(pst->mem, buffer, "gs_set buffer");
+
+    return code;
+}
+
+
 /* parse and set up state for one line of pjl commands */
 static int
 pjl_parse_and_process_line(pjl_parser_state_t * pst)
@@ -837,10 +904,10 @@ pjl_parse_and_process_line(pjl_parser_state_t * pst)
     token = (char *)gs_alloc_bytes(pst->mem, bufsize, "working buffer for PJL parsing");
     if (token == 0)
         return -1;
-    
+
     memset(token, 0x00, bufsize);
     memset(pathname, 0x00, sizeof(pathname));
-           
+
     /* reset the line position to the beginning of the line */
     pst->pos = 0;
     /* all pjl commands start with the pjl prefix @PJL */
@@ -872,8 +939,7 @@ pjl_parse_and_process_line(pjl_parser_state_t * pst)
 
                             strcpy(variable, token);
                             if (((tok = pjl_get_token(pst, token)) == EQUAL)
-                                && (tok =
-                                    pjl_get_token(pst, token)) == SETTING) {
+                                && (tok = pjl_get_token(pst, token)) == SETTING) {
                                 /* an unfortunate side effect - we have to
                                    identify FORMLINES being set here, for
                                    the benefit of PCL, see the
@@ -885,18 +951,46 @@ pjl_parse_and_process_line(pjl_parser_state_t * pst)
                                             (char *)"on", defaults);
                                 code = pjl_set(pst, variable, token,
                                                defaults);
-                                gs_free_object(pst->mem, variable, "2nd working buffer for PJL parsing");
-                                gs_free_object(pst->mem, token, "working buffer for PJL parsing");
-                                return code;
-                            } else {
-                                gs_free_object(pst->mem, variable, "2nd working buffer for PJL parsing");
-                                gs_free_object(pst->mem, token, "working buffer for PJL parsing");
-                                return -1;      /* syntax error */
-                            }
+                            } else
+                                code = -1; /* syntax error */
+                            gs_free_object(pst->mem, variable, "2nd working buffer for PJL parsing");
+                            gs_free_object(pst->mem, token, "working buffer for PJL parsing");
+                            return code;
                         } else
                             continue;
+
+                    gs_free_object(pst->mem, token, "working buffer for PJL parsing");
                     return 0;
                 }
+            case GSSET:
+            case GSSETSTRING:
+                {
+                    bool string;
+
+                    string = (tok == GSSETSTRING);
+                    code = 0;
+                    while ((tok = pjl_get_token(pst, token)) != DONE)
+                        if (tok == VARIABLE || tok == SETTING) {
+                            int len1 = strlen(token)+1;
+                            char *variable = (char *)gs_alloc_bytes(pst->mem, len1,
+                                                                    "2nd working buffer for PJL parsing");
+                            if (variable == NULL)
+                                code = -1;
+                            else {
+                                strcpy(variable, token);
+                                if (((tok = pjl_get_token(pst, token)) == EQUAL) &&
+                                    ((tok = pjl_get_token(pst, token)) == SETTING))
+                                    code = gs_set(pst, variable, token, string);
+                                else
+                                    code = gs_set(pst, variable, NULL, string);
+                                gs_free_object(pst->mem, variable, "2nd working buffer for PJL parsing");
+                            }
+                            break;
+                        }
+
+                    gs_free_object(pst->mem, token, "working buffer for PJL parsing");
+                    return code;
+                }
             case INITIALIZE:
                 /* set the user default environment to the factory default environment */
                 free_pjl_defaults(pst->mem, &pst->defaults);
@@ -1048,7 +1142,7 @@ pjl_parse_and_process_line(pjl_parser_state_t * pst)
 }
 
 /* get a file from 0: or 1: volume */
-static FILE *
+static gp_file *
 get_fp(pjl_parser_state_t * pst, char *name)
 {
     char result[MAXPATHLEN];
@@ -1062,7 +1156,7 @@ get_fp(pjl_parser_state_t * pst, char *name)
         if (result[0] == '\0')
             return 0;
     }
-    return gp_fopen(result, gp_fmode_rb);
+    return gp_fopen(pst->mem, result, gp_fmode_rb);
 }
 
 /* scan for a named resoource in the pcl sandbox 0: or 1: and return
@@ -1072,47 +1166,57 @@ long int
 pjl_get_named_resource_size(pjl_parser_state_t * pst, char *name)
 {
     long int size;
-    FILE *fp = get_fp(pst, name);
+    gp_file *fp = get_fp(pst, name);
 
     if (fp == NULL)
         return 0;
-    fseek(fp, 0L, SEEK_END);
-    size = ftell(fp);
-    fclose(fp);
+    size = gp_fseek(fp, 0L, SEEK_END);
+    if (size >= 0)
+        size = gp_ftell(fp);
+    else
+        size = 0;
+    gp_fclose(fp);
     return size;
 }
 
 /* get the contents of a file on 0: or 1: and return the result in the
    client allocated memory */
 int
-pjl_get_named_resource(pjl_parser_state * pst, char *name, byte * data)
+pjl_get_named_resource(pjl_parser_state * pst, char *name, byte * data, long int datasize)
 {
     long int size;
     int code = 0;
-    FILE *fp = get_fp(pst, name);
+    gp_file *fp = get_fp(pst, name);
 
     if (fp == NULL)
         return 0;
-    fseek(fp, 0L, SEEK_END);
-    size = ftell(fp);
-    rewind(fp);
-    if (size < 0 || (size != fread(data, 1, size, fp))) {
+    size = gp_fseek(fp, 0L, SEEK_END);
+    if (size >= 0)
+        size = gp_ftell(fp);
+    gp_rewind(fp);
+    if (size < 0 || size != datasize || (size != gp_fread(data, 1, size, fp))) {
         code = -1;
     }
-    fclose(fp);
+    gp_fclose(fp);
     return code;
 }
 
 /* set the initial environment to the default environment, this should
    be done at the beginning of each job */
-void
+int
 pjl_set_init_from_defaults(pjl_parser_state_t * pst)
 {
-    free_pjl_environment(pst->mem, &pst->envir);
-    set_pjl_environment(pst->mem, &pst->envir, pst->defaults);
+    int code = free_pjl_environment(pst->mem, &pst->envir);
+    if (code < 0)
+        return code;
+    code = set_pjl_environment(pst->mem, &pst->envir, pst->defaults);
+    if (code < 0)
+        return code;
 
-    free_pjl_fontsource(pst->mem, &pst->font_envir);
-    set_pjl_fontsource(pst->mem, &pst->font_envir, pst->font_defaults);
+    code = free_pjl_fontsource(pst->mem, &pst->font_envir);
+    if (code < 0)
+        return code;
+    return set_pjl_fontsource(pst->mem, &pst->font_envir, pst->font_defaults);
 }
 
 void
@@ -1180,8 +1284,10 @@ pjl_set_envvar(pjl_parser_state * pst, const char *pjl_var, const char *data)
             if (env[i].value)
                 gs_free_object(pst->mem, env[i].value, "pjl_set_envvar value");
             newvalue = (char *)gs_alloc_bytes(pst->mem, strlen(data) + 1, "pjl_set_envvar, value");
-            strcpy(newvalue, data);
-            env[i].value = newvalue;
+            if (newvalue != NULL) {
+                strcpy(newvalue, data);
+                env[i].value = newvalue;
+            }
         }
         i++;
     }
@@ -1203,8 +1309,10 @@ pjl_set_defvar(pjl_parser_state * pst, const char *pjl_var, const char *data)
             if (env[i].value)
                 gs_free_object(pst->mem, env[i].value, "pjl_set_defvar value");
             newvalue = (char *)gs_alloc_bytes(pst->mem, strlen(data) + 1, "pjl_set_defvar, value");
-            strcpy(newvalue, data);
-            env[i].value = newvalue;
+            if (newvalue != NULL) {
+                strcpy(newvalue, data);
+                env[i].value = newvalue;
+            }
         }
         i++;
     }
@@ -1252,15 +1360,43 @@ pjl_process(pjl_parser_state * pst, void *pstate, stream_cursor_read * pr)
                 /* Skip the UEL and continue. */
                 p += 9;
                 continue;
-            } else if (!memcmp(p + 1, "@PJL", min(avail, 4))) { /* Might be PJL. */
-                if (avail < 4) {        /* Not enough data to know yet. */
+            } else {
+                /* We allow for a single CRLF at the start of a block of PJL.
+                 * This is in violation of the spec, but allows us to accept
+                 * files like bug 693269. This shouldn't adversely affect any
+                 * real world files. */
+                if (avail > 0 && p[1] == '\r')
+                    p++, avail--;
+                if (avail > 0 && p[1] == '\n')
+                    p++, avail--;
+                if (!memcmp(p + 1, "@PJL", min(avail, 4))) { /* Might be PJL. */
+                    if (avail < 4) {        /* Not enough data to know yet. */
+                        break;
+                    }
+                    /* Definitely a PJL command. */
+                } else {            /* Definitely not PJL. */
+                    code = 1;
                     break;
                 }
-                /* Definitely a PJL command. */
-            } else {            /* Definitely not PJL. */
-                code = 1;
+            }
+        }
+        if (!legal_pjl_char(p[1])) {
+            code = 1;
+            /* If we haven't read anything yet, then just give up */
+            if (pst->pos == 0) {
+                ++p;
                 break;
             }
+            /* If we have, process what we had collected already before giving up. */
+            /* null terminate, parse and set the pjl state */
+            pst->line[pst->pos] = '\0';
+            /*
+             * NB PJL continues not considering normal errors but we
+             * ignore memory failure errors here and shouldn't.
+             */
+            (void)pjl_parse_and_process_line(pst);
+            pst->pos = 0;
+            break;
         }
         if (p[1] == '\n') {
             ++p;
@@ -1274,22 +1410,18 @@ pjl_process(pjl_parser_state * pst, void *pstate, stream_cursor_read * pr)
             pst->pos = 0;
             continue;
         }
-        
-        if (!legal_pjl_char(p[1])) {
-            code = 1;
-            ++p;
-            break;
-        }
 
         /* Copy the PJL line into the parser's line buffer. */
         /* Always leave room for a terminator. */
         if (pst->pos == pst->line_size) {
             char *temp = (char *)gs_alloc_bytes (pst->mem, pst->line_size + 256, "pjl_state increase line buffer");
-            memcpy(temp, pst->line, pst->line_size);
-            gs_free_object(pst->mem, pst->line, "pjl_state line buffer");
-            pst->line = temp;
-            pst->line_size += 256;
-            pst->line[pst->pos] = p[1], pst->pos++;
+            if (temp != NULL) {
+                memcpy(temp, pst->line, pst->line_size);
+                gs_free_object(pst->mem, pst->line, "pjl_state line buffer");
+                pst->line = temp;
+                pst->line_size += 256;
+                pst->line[pst->pos] = p[1], pst->pos++;
+            }
         } else
             pst->line[pst->pos] = p[1], pst->pos++;
         ++p;
@@ -1466,7 +1598,7 @@ static int set_pjl_environment(gs_memory_t * mem, pjl_envir_var_t **env, pjl_env
         limit++;
 
     pjl_env = (pjl_envir_var_t *) gs_alloc_bytes(mem,
-                                                  sizeof
+                                                  (size_t)sizeof
                                                   (pjl_envir_var_t) * (limit + 1),
                                                   "pjl_envir");
     if (!pjl_env)
@@ -1510,7 +1642,7 @@ static int set_pjl_fontsource(gs_memory_t * mem, pjl_fontsource_t **fontenv, pjl
         limit++;
 
     pjl_fontenv = (pjl_fontsource_t *) gs_alloc_bytes(mem,
-                                                      sizeof
+                                                      (size_t)sizeof
                                                       (pjl_fontsource_t) * (limit + 1),
                                                       "font_envir");
     if (!pjl_fontenv)
@@ -1563,7 +1695,7 @@ static int set_pjl_default_fontsource(gs_memory_t * mem, pjl_fontsource_t **font
         limit++;
 
     pjl_fontdef = (pjl_fontsource_t *) gs_alloc_bytes(mem,
-                                                      sizeof
+                                                      (size_t)sizeof
                                                       (pjl_fontsource_t) * (limit + 1),
                                                       "new_font_defaults");
     if (!pjl_fontdef)
@@ -1699,12 +1831,24 @@ pjl_process_init(gs_memory_t * mem)
         pathlen = 0;
         if ((code = gp_getenv("PCLFONTSOURCE", (char *)0, &pathlen)) < 0) {
             char *path =
-                (char *)gs_alloc_bytes(mem, pathlen, "pjl_font_path");
+                (char *)gs_alloc_bytes(mem, pathlen + 1, "pjl_font_path");
             /* if the allocation fails we use the pjl fontsource */
             if (path == NULL)
                 pjlstate->environment_font_path = NULL;
             else {
+                const char * const sepr = gp_file_name_separator();
+                const int lsepr = strlen(sepr);
                 gp_getenv("PCLFONTSOURCE", path, &pathlen);     /* can't fail */
+
+                /* We want to ensure a trailing "/" is present */
+                if (gs_file_name_check_separator(path + (pathlen - (lsepr + 1)), lsepr, path + pathlen - 1) != 1) {
+                    strncat(path, gp_file_name_separator(), pathlen + 1);
+                }
+                code = gs_add_control_path(mem, gs_permit_file_reading, path);
+                if (code < 0) {
+                    gs_free_object(mem, path, "pjl_font_path");
+                    goto fail1;
+                }
                 pjlstate->environment_font_path = path;
             }
         } else                  /* environmet variable does not exist use pjl fontsource */
@@ -1744,8 +1888,8 @@ pjl_process_init(gs_memory_t * mem)
     {
         int i;
 
-        for (i = 0; i < countof(pjl_permanent_soft_fonts); i++)
-            pjl_permanent_soft_fonts[i] = 0;
+        for (i = 0; i < countof(pjlstate->permanent_soft_fonts); i++)
+            pjlstate->permanent_soft_fonts[i] = 0;
     }
     return (pjl_parser_state *) pjlstate;
 
@@ -1781,7 +1925,7 @@ pjl_process_destroy(pjl_parser_state * pst)
         return;
 
     mem = pst->mem;
-    
+
     free_pjl_defaults(mem, &pst->defaults);
     free_pjl_environment(mem, &pst->envir);
     free_pjl_fontsource(mem, &pst->font_envir);
@@ -1804,10 +1948,10 @@ pjl_register_permanent_soft_font_deletion(pjl_parser_state * pst,
         return 0;
     }
     /* if the font is present. */
-    if ((pjl_permanent_soft_fonts[font_number >> 3]) &
+    if ((pst->permanent_soft_fonts[font_number >> 3]) &
         (128 >> (font_number & 7))) {
         /* set the bit to zero to indicate the fontnumber has been deleted */
-        pjl_permanent_soft_fonts[font_number >> 3] &=
+        pst->permanent_soft_fonts[font_number >> 3] &=
             ~(128 >> (font_number & 7));
         /* if the current font source is 'S' and the current font number
            is the highest number, and *any* soft font was deleted or if
@@ -1824,7 +1968,7 @@ pjl_register_permanent_soft_font_deletion(pjl_parser_state * pst,
             /* check for no more fonts and the highest font number.
                NB should look at longs not bits in the loop */
             for (i = 0; i < MAX_PERMANENT_FONTS; i++)
-                if ((pjl_permanent_soft_fonts[i >> 3]) & (128 >> (i & 7))) {
+                if ((pst->permanent_soft_fonts[i >> 3]) & (128 >> (i & 7))) {
                     empty = false;
                     highest_fontnumber = i;
                 }
@@ -1852,7 +1996,7 @@ pjl_register_permanent_soft_font_addition(pjl_parser_state * pst)
 
     for (font_num = 0; font_num < MAX_PERMANENT_FONTS; font_num++)
         if (!
-            ((pjl_permanent_soft_fonts[font_num >> 3]) &
+            ((pst->permanent_soft_fonts[font_num >> 3]) &
              (128 >> (font_num & 7)))) {
             slot_found = true;
             break;
@@ -1865,6 +2009,6 @@ pjl_register_permanent_soft_font_addition(pjl_parser_state * pst)
         font_num = 0;
     }
     /* set the bit to 1 to indicate the fontnumber has been added */
-    pjl_permanent_soft_fonts[font_num >> 3] |= (128 >> (font_num & 7));
+    pst->permanent_soft_fonts[font_num >> 3] |= (128 >> (font_num & 7));
     return font_num;
 }
