@@ -1,14 +1,16 @@
-/* Copyright (C) 2001-2012 Artifex Software, Inc.
+/* Copyright (C) 2001-2025 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
    implied.
 
-   This software is distributed under license and may not be copied, modified
-   or distributed except as expressly authorized under the terms of that
-   license.  Refer to licensing information at http://www.artifex.com/
-   or contact Artifex Software, Inc.,  7 Mt. Lassen Drive - Suite A-134,
-   San Rafael, CA  94903, U.S.A., +1(415)492-9861, for further information.
+   This software is distributed under license and may not be copied,
+   modified or distributed except as expressly authorized under the terms
+   of the license contained in the file LICENSE in this distribution.
+
+   Refer to licensing information at http://www.artifex.com or contact
+   Artifex Software, Inc.,  39 Mesa Street, Suite 108A, San Francisco,
+   CA 94129, USA, for further information.
 */
 
 /* Font API support  */
@@ -61,19 +63,18 @@ static const char *UFSTPLUGINPARM = "UFST_PlugIn=";
 extern const char gp_file_name_list_separator;
 
 /* forward declarations for the pl_ff_stub definition */
-static ulong
-pl_fapi_get_long(gs_fapi_font * ff, gs_fapi_font_feature var_id, int index);
+static int
+pl_fapi_get_long(gs_fapi_font * ff, gs_fapi_font_feature var_id, int index, unsigned long *ret);
 
 static int
-pl_fapi_get_cid(gs_font_base * pbfont, gs_string * charstring,
-                gs_string * name, int ccode, gs_string * enc_char_name,
+pl_fapi_get_cid(gs_text_enum_t *penum, gs_font_base * pbfont, gs_string * charstring,
+                gs_string * name, gs_glyph ccode, gs_string * enc_char_name,
                 char *font_file_path, gs_fapi_char_ref * cr, bool bCID);
 
 static int
-pl_fapi_get_glyph(gs_fapi_font * ff, int char_code, byte * buf,
-                  ushort buf_length);
+pl_fapi_get_glyph(gs_fapi_font * ff, gs_glyph char_code, byte * buf, int buf_length);
 
-static ushort
+static int
 pl_fapi_serialize_tt_font(gs_fapi_font * ff, void *buf, int buf_size);
 
 static int
@@ -82,12 +83,12 @@ pl_get_glyphdirectory_data(gs_fapi_font * ff, int char_code,
 
 static int
 pl_fapi_set_cache(gs_text_enum_t * penum, const gs_font_base * pbfont,
-                  const gs_string * char_name, int cid,
+                  const gs_string * char_name, gs_glyph cid,
                   const double pwidth[2], const gs_rect * pbbox,
                   const double Metrics2_sbw_default[4], bool * imagenow);
 
 static int
-pl_fapi_get_metrics(gs_fapi_font * ff, gs_string * char_name, int cid,
+pl_fapi_get_metrics(gs_fapi_font * ff, gs_string * char_name, gs_glyph cid,
                     double *m, bool vertical);
 
 static const gs_fapi_font pl_ff_stub = {
@@ -105,6 +106,7 @@ static const gs_fapi_font pl_ff_stub = {
     false,                      /* is_vertical */
     false,                      /* metrics_only */
     {{3, 10}, {3, 1}, {-1, -1}, {-1, -1}, {-1, -1}},    /* ttf_cmap_req */
+    {-1 , -1},                                          /* ttf_cmap_selected */
     0,                          /* client_ctx_p */
     0,                          /* client_font_data */
     0,                          /* client_font_data2 */
@@ -121,6 +123,7 @@ static const gs_fapi_font pl_ff_stub = {
     NULL,                       /* get_raw_subr */
     pl_fapi_get_glyph,          /* get_glyph */
     pl_fapi_serialize_tt_font,  /* serialize_tt_font */
+    NULL,                       /* retrieve_tt_font */
     NULL,                       /* get_charstring */
     NULL,                       /* get_charstring_name */
     pl_get_glyphdirectory_data, /* get_GlyphDirectory_data_ptr */
@@ -129,25 +132,28 @@ static const gs_fapi_font pl_ff_stub = {
     pl_fapi_set_cache           /* fapi_set_cache */
 };
 
-static ulong
-pl_fapi_get_long(gs_fapi_font * ff, gs_fapi_font_feature var_id, int index)
+static int
+pl_fapi_get_long(gs_fapi_font * ff, gs_fapi_font_feature var_id, int index, unsigned long *ret)
 {
     gs_font *pfont = (gs_font *) ff->client_font_data;
     pl_font_t *plfont = (pl_font_t *) pfont->client_data;
-    ulong value = -1;
+    int code = 0;
     (void)index;
 
     if (var_id == gs_fapi_font_feature_TT_size) {
-        value =
-            plfont->header_size - (plfont->offsets.GT +
+        *ret = plfont->header_size - (plfont->offsets.GT +
                                    (plfont->large_sizes ? 6 : 4));
     }
-    return (value);
+    else {
+        *ret = 0;
+        code = gs_note_error(gs_error_undefined);
+    }
+    return code;
 }
 
 static int
-pl_fapi_get_cid(gs_font_base * pbfont, gs_string * charstring,
-                gs_string * name, int ccode, gs_string * enc_char_name,
+pl_fapi_get_cid(gs_text_enum_t *penum, gs_font_base * pbfont, gs_string * charstring,
+                gs_string * name, gs_glyph ccode, gs_string * enc_char_name,
                 char *font_file_path, gs_fapi_char_ref * cr, bool bCID)
 {
     pl_font_t *plfont = pbfont->client_data;
@@ -157,11 +163,12 @@ pl_fapi_get_cid(gs_font_base * pbfont, gs_string * charstring,
     (void)enc_char_name;
     (void)font_file_path;
     (void)bCID;
+    (void)penum;
 
     if (plfont->allow_vertical_substitutes) {
         vertical = pl_font_vertical_glyph(ccode, plfont);
 
-        if (vertical != gs_no_glyph)
+        if (vertical != GS_NO_GLYPH)
             index = vertical;
     }
     cr->char_codes[0] = index;
@@ -169,8 +176,7 @@ pl_fapi_get_cid(gs_font_base * pbfont, gs_string * charstring,
 }
 
 static int
-pl_fapi_get_glyph(gs_fapi_font * ff, int char_code, byte * buf,
-                  ushort buf_length)
+pl_fapi_get_glyph(gs_fapi_font * ff, gs_glyph char_code, byte * buf, int buf_length)
 {
     gs_font *pfont = (gs_font *) ff->client_font_data;
     /* Zero is a valid size for a TTF glyph, so init to that.
@@ -189,12 +195,12 @@ pl_fapi_get_glyph(gs_fapi_font * ff, int char_code, byte * buf,
     return (size);
 }
 
-static ushort
+static int
 pl_fapi_serialize_tt_font(gs_fapi_font * ff, void *buf, int buf_size)
 {
     gs_font *pfont = (gs_font *) ff->client_font_data;
     pl_font_t *plfont = (pl_font_t *) pfont->client_data;
-    short code = -1;
+    int code = -1;
     int offset = (plfont->offsets.GT + (plfont->large_sizes ? 6 : 4));
 
     if (buf_size >= (plfont->header_size - offset)) {
@@ -202,7 +208,7 @@ pl_fapi_serialize_tt_font(gs_fapi_font * ff, void *buf, int buf_size)
 
         memcpy(buf, (plfont->header + offset), buf_size);
     }
-    return ((ushort) code);
+    return (code);
 }
 
 static int
@@ -213,7 +219,7 @@ pl_get_glyphdirectory_data(gs_fapi_font * ff, int char_code,
 }
 
 static int
-pl_fapi_get_metrics(gs_fapi_font * ff, gs_string * char_name, int cid,
+pl_fapi_get_metrics(gs_fapi_font * ff, gs_string * char_name, gs_glyph cid,
                     double *m, bool vertical)
 {
     gs_font_base *pfont = (gs_font_base *) ff->client_font_data;
@@ -242,11 +248,11 @@ pl_fapi_get_metrics(gs_fapi_font * ff, gs_string * char_name, int cid,
 
 static int
 pl_fapi_set_cache(gs_text_enum_t * penum, const gs_font_base * pbfont,
-                  const gs_string * char_name, int cid,
+                  const gs_string * char_name, gs_glyph cid,
                   const double pwidth[2], const gs_rect * pbbox,
                   const double Metrics2_sbw_default[4], bool * imagenow)
 {
-    gs_state *pgs = (gs_state *) penum->pis;
+    gs_gstate *pgs = penum->pgs;
     float w2[6];
     int code = 0;
     gs_fapi_server *I = pbfont->FAPI;
@@ -323,7 +329,7 @@ pl_fapi_set_cache(gs_text_enum_t * penum, const gs_font_base * pbfont,
 
 static int
 pl_fapi_set_cache_rotate(gs_text_enum_t * penum, const gs_font_base * pbfont,
-                         const gs_string * char_name, int cid,
+                         const gs_string * char_name, gs_glyph cid,
                          const double pwidth[2], const gs_rect * pbbox,
                          const double Metrics2_sbw_default[4],
                          bool * imagenow)
@@ -334,7 +340,7 @@ pl_fapi_set_cache_rotate(gs_text_enum_t * penum, const gs_font_base * pbfont,
 
 
 static int
-pl_fapi_build_char(gs_show_enum * penum, gs_state * pgs, gs_font * pfont,
+pl_fapi_build_char(gs_show_enum * penum, gs_gstate * pgs, gs_font * pfont,
                    gs_char chr, gs_glyph glyph)
 {
     int code;
@@ -393,7 +399,7 @@ pl_fapi_ufst_get_font_dir(gs_memory_t * mem)
     return (UFSTFONTDIR);
 }
 
-static void
+static int
 pl_get_server_param(gs_fapi_server * I, const char *subtype,
                     char **server_param, int *server_param_size)
 {
@@ -420,6 +426,7 @@ pl_get_server_param(gs_fapi_server * I, const char *subtype,
         *server_param = NULL;
         *server_param_size = length;
     }
+    return 1;
 }
 
 
@@ -482,14 +489,16 @@ pl_show_text_is_width_only(const gs_text_enum_t *pte)
 
 void pl_text_release(gs_text_enum_t *pte, client_name_t cname)
 {
-  (void)pte;
-  (void)cname;
+   gx_cpath_free((gx_clip_path *)pte->pcpath, "gx_default_text_release");
+   pte->pcpath = NULL;
+   rc_decrement_only(pte->dev, cname);
+   rc_decrement_only(pte->imaging_dev, cname);
   return;
 }
 
 static int
 pl_fapi_set_cache_metrics(gs_text_enum_t * penum, const gs_font_base * pbfont,
-                         const gs_string * char_name, int cid,
+                         const gs_string * char_name, gs_glyph cid,
                          const double pwidth[2], const gs_rect * pbbox,
                          const double Metrics2_sbw_default[4],
                          bool * imagenow)
@@ -511,10 +520,10 @@ pl_fapi_char_metrics(const pl_font_t * plfont, const void *vpgs,
     gs_font_base *pbfont = (gs_font_base *) pfont;
     gs_text_params_t text;
     gs_char buf[2];
-    gs_state *rpgs = (gs_state *) vpgs;
+    gs_gstate *rpgs = (gs_gstate *) vpgs;
     /* NAFF: undefined glyph would be better handled inside FAPI */
     gs_char chr = char_code;
-    gs_glyph unused_glyph = gs_no_glyph;
+    gs_glyph unused_glyph = GS_NO_GLYPH;
     gs_glyph glyph;
     gs_matrix mat = {72.0, 0.0, 0.0, 72.0, 0.0, 0.0};
     gs_matrix fmat;
@@ -529,24 +538,23 @@ pl_fapi_char_metrics(const pl_font_t * plfont, const void *vpgs,
     if (pfont->WMode & 1) {
         gs_glyph vertical = pl_font_vertical_glyph(glyph, plfont);
 
-        if (vertical != gs_no_glyph)
+        if (vertical != GS_NO_GLYPH)
             glyph = vertical;
     }
 
     /* undefined character */
-    if (glyph == 0xffff || glyph == gs_no_glyph) {
+    if (glyph == 0xffff || glyph == GS_NO_GLYPH) {
         metrics[0] = metrics[1] = metrics[2] = metrics[3] = 0;
         code = 1;
     } else {
         gs_fapi_server *I = pbfont->FAPI;
-        gs_state lpgs;
-        gs_state *pgs = &lpgs;
+        gs_gstate lpgs;
+        gs_gstate *pgs = &lpgs;
 
         /* This is kind of naff, but it's *much* cheaper to copy
          * the parts of the gstate we need, than gsave/grestore
          */
         memset(pgs, 0x00, sizeof(lpgs));
-        pgs->is_gstate = rpgs->is_gstate;
         pgs->memory = rpgs->memory;
         pgs->ctm = rpgs->ctm;
         pgs->in_cachedevice = CACHE_DEVICE_NOT_CACHING;
@@ -571,12 +579,12 @@ pl_fapi_char_metrics(const pl_font_t * plfont, const void *vpgs,
         text.data.chars = buf;
         text.size = 1;
 
-        if ((penum1 = gs_text_enum_alloc(pfont->memory, (gs_imager_state *)pgs,
+        if ((penum1 = gs_text_enum_alloc(pfont->memory, pgs,
                       "pl_fapi_char_metrics")) != NULL) {
 
             if ((code = gs_text_enum_init(penum1, &null_text_procs,
-                                 NULL, (gs_imager_state *)pgs, &text, pfont,
-                                 NULL, NULL, NULL, pfont->memory)) >= 0) {
+                                 NULL, pgs, &text, pfont,
+                                 NULL, pfont->memory)) >= 0) {
 
                 code = gs_fapi_do_char(pfont, pgs, penum1, plfont->font_file, false,
                             NULL, NULL, char_code, glyph, 0);
@@ -648,18 +656,10 @@ pl_fapi_passfont(pl_font_t * plfont, int subfont, char *fapi_request,
     code =
         gs_fapi_passfont(pfont, subfont, (char *)file_name, &fdata,
                          (char *)fapi_request, NULL, (char **)&fapi_id,
-                         (gs_fapi_get_server_param_callback)
-                         pl_get_server_param);
+                         NULL, pl_get_server_param);
 
-    if (pfont->FontType == ft_MicroType && code < 0) {
-        return (code);
-    }
-
-    /* For now, we'll fall back to AFS
-       We should return an error in the future.
-     */
     if (code < 0 || fapi_id == NULL) {
-        return (0);
+        return code;
     }
 
     pfont->procs.build_char = pl_fapi_build_char;
@@ -676,9 +676,7 @@ bool
 pl_fapi_ufst_available(gs_memory_t * mem)
 {
     gs_fapi_server *serv = NULL;
-    int code = gs_fapi_find_server(mem, (char *)"UFST", &serv,
-                                   (gs_fapi_get_server_param_callback)
-                                   pl_get_server_param);
+    int code = gs_fapi_find_server(mem, (char *)"UFST", &serv, pl_get_server_param);
 
     if (code == 0 && serv != NULL) {
         return (true);
